// MiniSAX.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "StringHolder.h"
#include "CharBuffer.h"
#include "MiniSAX.h"


#define DEFINED(a) ((a) != NULL)


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ulReasonForCall, 
                       LPVOID lpReserved
					 )
{
    switch (ulReasonForCall)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}


using namespace miniSAX;


/*****************************
  CIOException implementation
 *****************************/


CIOException::CIOException(const char* aMessage) {
	if (DEFINED(aMessage)) {
		message = new char[strlen(aMessage) + 1];
		strcpy( message, aMessage );
	}
	else {
		message = NULL;
	}
}


CIOException::~CIOException() {
	if (DEFINED(message)) {
		delete [] message;
		message = NULL;
	}
}


/****************************
  CDataSource implementation
 ****************************/


CDataSource::CDataSource(const char* fileName) throw (CIOException) {
	loadFromFile( fileName );
}


CDataSource::~CDataSource() {
	freeData();
}


void CDataSource::freeData() {
	if (DEFINED(data)) {
		free(data);
		len    = 0;
		offset = 0;
		col    = 1;
		row    = 1;
	}
}


void CDataSource::loadFromFile(const char* fileName)
	throw (CIOException)
{
	const size_t	BUFFER_SIZE = 32*1024;

	// set startup data buffer
	size_t	bufferSize = BUFFER_SIZE;
	data   = (char*)malloc( bufferSize );
	len    = 0;
	offset = 0;
	col    = 1;
	row    = 1;

	// open file
	HANDLE	fileHandle = ::CreateFile( fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL );
	if (fileHandle == INVALID_HANDLE_VALUE) {
		throw CIOException( "Failed to open file for reading" );
	}

	// prepare read buffer
	char*	buffer = (char*)malloc( BUFFER_SIZE );

	// read file
	DWORD	bytesRead;
	do {
		if (!::ReadFile( fileHandle, buffer, BUFFER_SIZE, &bytesRead, NULL )) {
			break;
		}
		if (bytesRead > 0) {
			// check data size that can be used
			if ((bufferSize - len) < bytesRead) {
				size_t	needed = len + bytesRead;
				bufferSize <<= 1;
				while (bufferSize < needed) {
					bufferSize <<= 1;
				}
				char*	newData = (char*)realloc( data, bufferSize );
				if (!DEFINED(newData)) {
					// TODO: handle error
					::CloseHandle( fileHandle );
					return;
				}
				data = newData;
			}
			memcpy( data + len, buffer, bytesRead );
			len += bytesRead;
		}
	} while (bytesRead > 0);

	// free read buffer
	free( buffer );

	// close file
	::CloseHandle( fileHandle );
}


bool CDataSource::hasData() const {
	return (offset < len);
}


char CDataSource::getChar() {
	// TODO: check offset (data presence)
	char	currentChar = data[offset++];
	if (currentChar == '\n') {
		col++;
		row=1;
	}
	else {
		row++;
	}
	return currentChar;
}


void CDataSource::skipSpaces() {
	while (offset < len) {
		char	currentChar = data[offset];
		if (currentChar != ' ' && currentChar != '\n' && currentChar != '\r' && currentChar != '\t') {
			break;
		}
		if (currentChar == '\n') {
			col++;
			row=1;
		}
		else {
			row++;
		}
		offset++;
	}
}


size_t	CDataSource::getDistance(char delimiter) const {
	size_t	distance = offset;
	while (distance < len && data[distance] != delimiter) {
		distance++;
	}

	return (distance - offset);
}


size_t CDataSource::getNonElementDistance() const {
	size_t	distance = offset;
	while (distance < len) {
		char	currentChar = data[distance];
		if (isalpha(currentChar) || isdigit(currentChar) || currentChar == '_' || currentChar == '-') {
			distance++;
		}
		else {
			break;
		}
	}

	return (distance - offset);
}


void CDataSource::getSequence(char* dest, int seqLen) {
	// TODO: check boundaries
	if (seqLen > 0) {

		// move position
		int	from = offset;
		int	to   = offset + seqLen - 1;
		for (int i = from; i <= to; i++) {
			if (data[i] == '\n') {
				col++;
				row=1;
			}
			else {
				row++;
			}
		}

		memcpy( dest, data + offset, seqLen );
		offset += seqLen;
	}
}


void CDataSource::back() {
	offset--;
	if (data[offset] == '\n') {
		col--;
		// TODO: how to restore row value
	}
	else {
		row--;
	}
}


/*******************
  CMiniSAXException
 *******************/


CMiniSAXException::CMiniSAXException(const char* aMessage, int aCol, int aRow) {
	message = new char[strlen(aMessage)+1];
	strcpy( message, aMessage );
	col = aCol; row = aRow;
}


CMiniSAXException::CMiniSAXException(const char* aMessage) {
	message = new char[strlen(aMessage)+1];
	strcpy( message, aMessage );
	col = -1; row = -1;
}


CMiniSAXException::~CMiniSAXException() {
	delete [] message;
}


/************
  CAttribute
 ************/


CAttribute::CAttribute() {
	name = NULL;
	value = NULL;
}


CAttribute::CAttribute(const char* aName, const char* aValue) {
	store( aName, aValue );
}


CAttribute::CAttribute(const CAttribute& instance) {
	store( instance.getName(), instance.getValue() );
}


CAttribute::~CAttribute() {
	freeMem();
}


void CAttribute::store(const char* aName, const char* aValue) {
	if (DEFINED(aName)) {
		name = new char[strlen(aName)+1];
		strcpy( name, aName );
	}
	else {
		name = NULL;
	}

	if (DEFINED(aValue)) {
		value = new char[strlen(aValue)+1];
		strcpy( value, aValue );
	}
	else {
		value = NULL;
	}
}


void CAttribute::freeMem() {
	if (DEFINED(name)) {
		delete [] name;
		name = NULL;
	}

	if (DEFINED(value)) {
		delete [] value;
		value = NULL;
	}
}


CAttribute& CAttribute::operator =(const CAttribute& instance) {
	freeMem();
	store( instance.getName(), instance.getValue() );

	return (*this);
}


/****************
  CAttributeList
 ****************/


CAttributeList::CAttributeList() {
	attributes = NULL;
	count      = 0;
}


CAttributeList::~CAttributeList() {
	if (DEFINED(attributes))
		delete [] attributes;
}


void CAttributeList::addAttribute(const CAttribute& newAttr) {
	CAttribute*	newAttrs = new CAttribute[count+1];
	for (int i = 0; i < count; i++) {
		newAttrs[i] = attributes[i];
	}
	newAttrs[count] = newAttr;
	count++;
	if (DEFINED(attributes)) {
		delete [] attributes;
	}
	attributes = newAttrs;
}


const CAttribute* CAttributeList::getAttribute(int index) const {
	// TODO: check boundaries
	return &(attributes[index]);
}


const CAttribute* CAttributeList::getAttribute(const char* name) const {
	for (int i = 0; i < count; i++) {
		if (strcmp( attributes[i].getName(), name ) == 0) {
			return &(attributes[i]);
		}
	}

	return NULL;
}


/*******************************
  CMiniSAXParser implementation
 *******************************/


CMiniSAXParser::CMiniSAXParser() {
	// reset all handlers
	documentHandler = NULL;
	contentHandler  = NULL;
	errorHandler    = NULL;
	entityResolver  = NULL;

	currentRow    = -1;
	currentColumn = -1;
}


CMiniSAXParser::~CMiniSAXParser()
{ 
}


void CMiniSAXParser::parse(CDataSource* dataSource) throw(CMiniSAXException)
{
	// prepare buffer
	CCharBuffer	buffer( 1024 );
	CCharBuffer	storage( 1024 );	// will be used to resolve entities

	try {
		// start document parsing
		startDocument();

		// first symbol should be '<'

		// remember position
		currentRow    = dataSource->getRow();
		currentColumn = dataSource->getColumn();

		// get position into local variables
		int	col = dataSource->getColumn();
		int	row = dataSource->getRow();

		if (!dataSource->hasData()) {
			handleError( CMiniSAXException( "unexpected end of data", col, row ) );
			return;
		}

		if (dataSource->getChar() != '<') {
			handleError( CMiniSAXException( "unexpected symbol", col, row ) );
			return;
		}

		if (!readStartOfElement(dataSource)) {
			return;
		}

		bool	noError = true;
		do {
			// remember position
			currentRow    = dataSource->getRow();
			currentColumn = dataSource->getColumn();

			int	distance = dataSource->getDistance( '<' );
			if (distance > 0) {
				// increase buffers if have to (+1 for '\0')
				buffer.checkSize( distance + 1 );
				storage.checkSize( distance + 1 );

				dataSource->getSequence( buffer, distance );
				buffer(distance) = '\0';

				resolveEntities( buffer, storage );

				// handle characters
				handleCharacters( buffer, strlen( buffer ) );
			}
			
			// remember position again
			currentRow    = dataSource->getRow();
			currentColumn = dataSource->getColumn();

			if (dataSource->hasData()) {
				// determine '<''/'
				bool	endOfElement = false;
				// read '<'
				dataSource->getChar();
				if (dataSource->hasData()) {
					if (dataSource->getChar() == '/') {
						endOfElement = true;
					}
					else {
						dataSource->back();
					}
				}
				if (endOfElement) {
					noError = readEndOfElement(dataSource);
				}
				else {
					noError = readStartOfElement(dataSource);
				}
			}
		} while (noError && dataSource->hasData());

		// end document parsing
		endDocument();

	} catch(...) {
		buffer.reset();

		throw;
	}
}


bool CMiniSAXParser::readStartOfElement(CDataSource* dataSource) {
	dataSource->skipSpaces();

	// read name

	char*	name;
	if (!readName( dataSource, name )) {
		return false;
	}

	// holder will release memory
	CStringHolder	nameHolder(name);

	// read attributes

	CAttributeList	attributes;

	bool	haveAllData  = false;
	bool	endOfElement = false;
	do {
		dataSource->skipSpaces();
		// check close bracket of element
		if (!dataSource->hasData()) {
			handleError( CMiniSAXException( "unexpected end of data", 
											dataSource->getColumn(),
											dataSource->getRow() ) );
			return false;
		}
		switch (dataSource->getChar()) {
		case '>':
			haveAllData = true;
			break;
		case '/':
			{
				int	col = dataSource->getColumn();
				int	row = dataSource->getRow();
				if (!dataSource->hasData()) {
					handleError( CMiniSAXException( "unexpected end of data", col, row ) );
					return false;
				}
				if (dataSource->getChar() != '>') {
					handleError( CMiniSAXException( "unexpected symbol", col, row ) );
					return false;
				}
				// it's the end of the element
				endOfElement = true;
				haveAllData = true;
			}
			break;
		default:
			dataSource->back();
		}

		if (!haveAllData) {
			// read attribute name
			char*	attrName;
			if (!readName( dataSource, attrName )) {
				return false;
			}

			// holder will release memory
			CStringHolder	attrNameHolder( attrName );

			dataSource->skipSpaces();

			// check '='
			int	col = dataSource->getColumn();
			int	row = dataSource->getRow();
			if (!dataSource->hasData()) {
				handleError( CMiniSAXException( "unexpected end of data", col, row ) );
				return false;
			}
			if (dataSource->getChar() != '=') {
				handleError( CMiniSAXException( "unexpected symbol", col, row ) );
				return false;
			}

			dataSource->skipSpaces();

			// read attribute value
			char*	attrValue;
			if (!readValue( dataSource, attrValue )) {
				return false;
			}

			// holder will release memory
			CStringHolder	attrValueHolder( attrValue );

			// create attribute
			CAttribute	attr( attrName, attrValue );

			attributes.addAttribute( attr );
		}
	} while (!haveAllData);

	startElement( name, attributes );
	if (endOfElement) {
		endElement( name );
	}

	return true;
}


bool CMiniSAXParser::readEndOfElement(CDataSource* dataSource) {
	dataSource->skipSpaces();

	// read name

	char*	name;
	if (!readName( dataSource, name )) {
		return false;
	}

	// holder will release memory on exit
	CStringHolder	nameHolder(name);

	dataSource->skipSpaces();
	int	col = dataSource->getColumn();
	int	row = dataSource->getRow();
	if (!dataSource->hasData()) {
		handleError( CMiniSAXException( "unexpected end of data", col, row ) );
		return false;
	}
	if (dataSource->getChar() != '>') {
		handleError( CMiniSAXException( "unexpected symbol", col, row ) );
		return false;
	}

	endElement( name );

	return true;
}


bool CMiniSAXParser::readName(CDataSource* dataSource, char*& name) {
	int		col       = dataSource->getColumn();
	int		row       = dataSource->getRow();
	char	firstChar = dataSource->getChar();
	if (!isalpha(firstChar) && firstChar != '_') {
		handleError( CMiniSAXException( "unexpected symbol", col, row ) );
		return false;
	}

	size_t	distance = dataSource->getNonElementDistance();

	char*	buffer = new char[distance+2];	// + first char + '\0'
	if (distance > 0) {
		dataSource->getSequence( buffer + 1, distance );
	}
	buffer[0]          = firstChar;
	buffer[distance+1] = '\0';

	// we have name
	name = buffer;

	return true;
}


bool CMiniSAXParser::readValue(CDataSource* dataSource, char*& value) {
	int		col       = dataSource->getColumn();
	int		row       = dataSource->getRow();
	char	firstChar = dataSource->getChar();
	if (firstChar != '\'' && firstChar != '\"') {
		handleError( CMiniSAXException( "unexpected symbol", col, row ) );
		return false;
	}

	size_t	distance = dataSource->getDistance( firstChar );

	char*	buffer = new char[distance+1];	// + '\0'
	if (distance > 0) {
		dataSource->getSequence( buffer, distance );
	}
	buffer[distance] = '\0';

	char*	storage = new char[distance+1];
	resolveEntities( buffer, storage );
	delete [] storage;

	if (!dataSource->hasData()) {
		handleError( CMiniSAXException( "unexpected end of data", 
									    dataSource->getColumn(),
										dataSource->getRow() ) );
		return false;
	}

	// read closing quote
	dataSource->getChar();

	// we have value
	value = buffer;

	return true;
}


void CMiniSAXParser::startDocument() {
	if (DEFINED(documentHandler)) {
		documentHandler->startDocument();
	}
}


void CMiniSAXParser::endDocument() {
	if (DEFINED(documentHandler)) {
		documentHandler->endDocument();
	}
}


void CMiniSAXParser::startElement(const char* name, const CAttributeList& attributes) {
	if (DEFINED(contentHandler)) {
		contentHandler->startElement( name, attributes );
	}
}


void CMiniSAXParser::endElement(const char* name) {
	if (DEFINED(contentHandler)) {
		contentHandler->endElement( name );
	}
}


void CMiniSAXParser::handleCharacters(const char* buffer, size_t len) {
	if (DEFINED(contentHandler)) {
		contentHandler->handleCharacters( buffer, len );
	}
}



void CMiniSAXParser::handleError(const CMiniSAXException& e) {
	if (DEFINED(errorHandler)) {
		errorHandler->handleError( e );
	}
}


void CMiniSAXParser::resolveEntities(char* data, char* storage) {
	int	pos = 0;
	while (data[pos] != '\0') {
		if (data[pos] == '&') {
			++pos;
			if (strncmp( data + pos, "lt;", 3 ) == 0) {
				strcpy( storage, data + pos + 3 );
				data[pos-1] = '<';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "gt;", 3 ) == 0) {
				strcpy( storage, data + pos + 3 );
				data[pos-1] = '>';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "apos;", 5 ) == 0) {
				strcpy( storage, data + pos + 5 );
				data[pos-1] = '\'';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "quot;", 5 ) == 0) {
				strcpy( storage, data + pos + 5 );
				data[pos-1] = '"';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "amp;", 4 ) == 0) {
				strcpy( storage, data + pos + 4 );
				data[pos-1] = '&';
				strcpy( data+pos, storage );
				continue;
			}
			// TODO: we've met '&' but failed to determine entity, we should raise an exception
		} else {
			++pos;
		}
	}
}
